.check_periods <- function(x, periods) {
  idx <- time(na.omit(x))
  nobs <- sapply(periods, function(rng) nrow(x[rng, ]))

  lacking <- names(nobs[nobs < 10])
  return(lacking)
}

compare_lf <- function(x,
                       periods = c(y2015   = "2015-01-01::2015-12-31",
                                   y2003   = "2003-01-01::2003-12-31"),
                       refPeriod = "1976-01-01::2010-12-31",
                       dist.min = "wei", dist.max = "gev", prob = 0.2) {

  name <- xtsAttributes(x)$station
  vars.max <- c("duration", "volume", "intensity")
  vars.time <- c("start", "end", "time")  # they do not show up in the results

  vars <- c("discharge", vars.max, "day", vars.time)

  # concatenate the periods
  per <- c(periods, reference = refPeriod)

  lacking <- .check_periods(x, periods)
  if(length(lacking)) {
    warning(name, ": Not enough data for period(s): ",
            paste(sQuote(lacking), collapse = ", "),
            sep = "", call. = FALSE)
  }

  # skeleton objects for return period, observed values and mean of ref period
  values <- rp <- matrix(NA, nrow = length(per), ncol = length(vars),
                         dimnames = list(names(per), vars))

  # computing the threshold
  threshold <- quantile(as.vector(x[refPeriod, "discharge"]), probs = prob, na.rm = TRUE)

  # apply a 7-day moving average, NAs are introduced at bounds
  x$ma7 <- ma(as.vector(x$discharge), n = 7, sides = 2)

  # compute annual minima for the whole range, subsequently subset it for
  # for the cuurent year (eg. 2015), 2003 and for the reference period
  am7 <- .am(x$ma7, name = name)

  # bug! if qmin is outside of period, NA is returned
  dis <- lapply(per, function(x) as.data.frame(am7[x, ]))
  tmp <- do.call(rbind, lapply(dis, colMeans))
  colnames(tmp) <- c("discharge", "day")
  values[rownames(tmp), colnames(tmp)] <- as.matrix(tmp)

  # also store monthly minima
  # apply.monthly does not return the index of min, just the endpoint of the period
  mm7 <- suppressWarnings(apply.monthly(x$ma7, min, na.rm = T))

  # mean values for day of occurrence is computed differently
  values[names(per)[3], "day"] <-
    round(.circular_mean(dis[["reference"]]$day)[1], 1)

  # fit a weibull distribution for the discharges
  fitd <- suppressWarnings(evfit(dis[["reference"]]$ma7, distribution = dist.min))

  # return period of discharge
  rp[, "discharge"] <- .rp_ev(x = values[, "discharge"], fit = fitd)


  # find drought events
  drought <- pool_sp(find_droughts(x$discharge, threshold = threshold))

  # todo: do this for every period
  unterminated <- tail(na.omit(drought)$event.no, 1) != 0
  if(unterminated) {
    warning(name, ": Time series ended before recovery of (pooled) drought.", sep = "",
            call. = FALSE)
  }

  tab <- suppressWarnings(summary(drought, drop_minor = F))
  tab$intensity <- tab$volume / tab$duration
  rng <- as.numeric(format(range(time(x)), "%Y"))
  tab$year <- .year_as_factor(tab$time, from = rng[1], to = rng[2])

  .expand_drought <- function(x) {
    start <- as.numeric(format(x$start, "%Y"))
    end   <- as.numeric(format(x$end, "%Y"))
    yseq <- seq.int(start, end)

    y <- x[rep(1, length(yseq)), ]
    y$year <- yseq
    y$attributed <- as.numeric(as.character(x$year))

    return(y)
  }

  y <- list()
  for(i in seq_len(nrow(tab))) {
    y[[i]] <- .expand_drought(tab[i, ])
  }
  y <- do.call(rbind, y)
  rownames(y) <- NULL

  tab <- y

  # select event with maximum volume
  # set values for years without deficit to zero
  run <- by(tab[, c(vars.max, vars.time, "attributed")], INDICES = tab$year,
            function(x) x[which.max(x$volume), ])
  run[sapply(run, is.null)] <- 0
  run <- do.call(rbind, run)

  run[, "multiyear"] <- .year(run$end) - .year(run$start) + 1

  run[, "start"] <- as.numeric(run[, "start"])
  run[, "end"] <- as.numeric(run[, "end"])
  run[, "time"] <- as.numeric(run[, "time"])

  run.all <- xts(run, order.by = as.Date(paste0(rownames(run), "-01-01")))

  run <- lapply(per, function(x) as.data.frame(run.all[x, ]))
  tmp <- do.call(rbind, lapply(run, colMeans))

  # for ref.period:mean day of onset and end has to be a circular mean
  tmp["reference", vars.time] <- apply(run[["reference"]][, vars.time], 2, function(x)
    round(.circular_mean(as.numeric(format(as.Date(x, origin = "1970-01-01"), "%j")))[1], 1))

  values[rownames(tmp), c(vars.max, vars.time)] <- as.matrix(tmp[, c(vars.max, vars.time)])

  maxima <- run[["reference"]]
  maxima <- maxima[!duplicated(maxima$attributed), vars.max]
  fit <- lapply(maxima, function(x)
    suppressWarnings(evfit(as.vector(x), distribution = dist.max, extreme = "max")))

  time <- as.list(as.data.frame(t(values[, vars.time])))
  time[names(time) != "reference"] <- lapply(time[names(time) != "reference"],
                                             as.Date, origin = "1970-01-01")
  time <- as.data.frame(time, row.names = vars.time)[c("start", "end"), ]

  values <- values[, setdiff(colnames(values), vars.time)]
  rp <- rp[, setdiff(colnames(rp), vars.time)]

  for(j in names(fit)) rp[, j] <- .rp_ev(x = values[, j], fit[[j]])

  lvls <- c("terminated", "unterminated (but not most extreme)", "unterminated (and most extreme)")
  # check if "end" of drought is identical with end of time series
  # station is still under drought
  idx <- if(!unterminated) 1 else if(!is.na(time["end", 1]) && time["end", 1] == tail(time(x), 1)) 3 else 2
  terminated <- factor(lvls[idx], levels = lvls )

  res <- list(values = values,
              return.period = round(rp, 2),
              fit =  c(list(discharge = fitd), fit))

  class(res) <- c("hpt", "list")

  attr(res, "lf") <- c(xtsAttributes(x),
                       drought = list(run),
                       drought.tbl = list(run.all),
                       timing = list(time),
                       threshold = threshold,
                       refPeriod = refPeriod,
                       periods = list(periods),
                       prob = prob,
                       unterminated = unterminated,
                       terminated = list(terminated),
                       mm7 = list(mm7),
                       meta = list(list(date.of.analysis = Sys.Date(),
                                   version = list(drought = packageVersion("drought2015"),
                                                  lfstat = packageVersion("lfstat"),
                                                  R = R.Version()$version.string))))

  .check_attr(res)
  return(res)

}

print.hpt <- function(x, ...) {
  #print(lapply(x[c("values", "return.period")], signif, 3))
  print(x[c("values", "return.period")])
}


.format.day <- function(x, ref, suffix = " than ref") {
  d <- x - ref
  sign <- if (is.na(d)) "NA" else {
    if (d < 0) "earlier" else "later"
  }
  suffix <- paste0(sign, suffix)
  paste0("on day ", format(x, nsmall = 1, width = 5),
         " (", format(round(abs(d), 0), width = 3), " days ", suffix, ")")
}



summary.hpt <- function(object, ...) {
  x <- object

  att <- attr(x, "lf")
  coord <- format(round(att$coordinates, 2), big.mark = " ", nsmall = 2)

  cat("River ", sQuote(att$river), " at station ", sQuote(att$station),
      ", ", att$region, " (", att$country, ")\n", sep = "")
  cat("Catchment size: ", att$catchment, "km2\n")
  cat("Coordinates (x, y, z): ", paste(coord, collapse = ",  "), "\n")
  cat("Data coverage: ", paste(att$time, collapse = " to "),
      " (", round(att$data.coverage * 100, 1), "% missing values)", "\n", sep = "")
  if(att$unterminated) {
    cat("Time series ended before recovery of (pooled) drought.\n\n")
  }


  cat("Threshold: Q", (1-att$prob)*100, " = ", signif(att$threshold, 3),
      "m3/s\n\n", sep = "")


  val <- x[["values"]][, setdiff(colnames(x[["values"]]), "day")]

  cat("Absolute Values:\n")
  print(signif(x$values, digits = 3))

  cat("\nRelative Values of 2015 compared to\n")
  rel <- round(rep(val[1, ], each = nrow(val)-1) / val[-1, ], 2)
  print(rel)

  cat("\nDay of Occurrence of minimum Discharge:\n")
  day <- round(x$values[, "day"], 1)
  name <- paste(format(names(day)), "  ")

  n <- length(day)


  cat(name[1], " on day ", format(day[1], nsmall = 1, width = 5), " (",
      .date_english(day[1]), ")\n",
      sep = "")
  for(i in 2:n) {
    cat(name[i],  .format.day(day[i], day[1],
                              suffix = paste0(" than ", names(day[1]))), "\n")
  }

  cat("\n\nReturn Period in Years\n")
  print(x$return.period[, setdiff(colnames(x$return.period), "day")])

  cat("\n\nDrought timings\n")
  print(att$timing)

  cat("\n\nContingency Table for Multi-annual Droughts lasting n Years:\n")
  print(.multiyear(att$drought))
}

.multiyear <- function(x) {
  x <- x[sapply(x, nrow) > 0]
  m <- lapply(x, function(x) table(unique(x[, c("attributed", "multiyear")])$multiyear))

  ncol <- max(sapply(m, function(x)  max(as.numeric(names(x)))))
  tbl <- matrix(0, nrow = length(m), ncol = ncol,
                dimnames = list(Period = names(m), "lasting n Years" = seq_len(ncol)))
  for(per in seq_along(m)){
    row <- rep(per, length(m[[per]]))
    tbl[matrix(c(row, as.numeric(names(m[[per]]))), ncol = 2)] <- m[[per]]
  }

  return(tbl)
}


plot.hpt <- function(x, y = NULL, ...) {
  val <- x[["values"]][, setdiff(colnames(x[["values"]]), "day")]
  rp <- x[["return.period"]][, setdiff(colnames(x[["values"]]), "day")]

  oldpar <- par(no.readonly = TRUE)
  par(mfcol = c(2, 2), mar =  c(3, 4, 1, 2) + 0.1)

  n <- nrow(val)
  col <- c("#F8766D", "#B79F00", "#00BA38", "#00BFC4", "#619CFF", "#F564E3")
  col <- c(head(col, n-1), "black")

  .diagplot <- function(fit, var, prob, quantile, ylab, legend = T, col.rp = col){
    xlab <- ""
    plot(fit[[var]], log = F, ylab = ylab, xlab = xlab, rp.lab = "", legend = legend)
    trace_value(prob[, var], quantile[, var], col = col.rp, annotate = F)
  }

  .diagplot(fit = x$fit, prob = 1/rp, quantile = val, var = "discharge",
            ylab = "Discharge in m3/s")
  .diagplot(fit = x$fit, prob = 1 - 1/rp, quantile = val, var = "volume",
            ylab = "Deficit Volume in m3")
  .diagplot(fit = x$fit, prob = 1 - 1/rp, quantile = val, var = "duration",
            ylab = "Deficit Duration in days")
  .diagplot(fit = x$fit, prob = 1 - 1/rp, quantile = val, var = "intensity",
            ylab = "Deficit Intensity in m3/days", legend = FALSE)

  legend(x = "topleft", legend = rownames(val), col = col, lty = 2)
  par(oldpar)
}

.meta <- list(required = c("coordinates", "unit", "country", "filename", "epsg"),
              optional = c("station", "river", "region", "catchment"))

.check_attr <- function(x) {
  if(!length(x)) {
    warning("empty object passed on to .check_attr", call. = FALSE)
    return()
  } else if(inherits(x, what = "xts")){
    actual <- xtsAttributes(x)
  } else if(inherits(x, "hpt"))  {
    actual <- attr(x, "lf")
    .meta$required <- c(.meta$required, "mm7")
  } else {
    warning(".check_attr does not know how to handle an object of class ",
            class(x))
    return()
  }

  # check if relevant attributes exist
  missing <- setdiff(.meta$required, names(actual))
  if(length(missing)) {
    warning(.station(x), "Required argument ",
            paste(shQuote(missing), collapse = ", "), " does not exist.")
  }

  # TODO check if attributes are of right class

  # check if unit is correct
  if(actual$unit != "m^3/s") {
    warning(.station(x), "Unit has to be 'm^3/s'. ")
  }

}

.station <- function(x) {
  if(inherits(x, what = "xts")){
    actual <- xtsAttributes(x)
  } else if(inherits(x, "hpt"))  {
    actual <- attr(x, "lf")
  }

  txt <- with(actual, paste0(country, " (", filename, ")",
                             ifelse(length(station), paste0(" ", station, "")), ": "))

  return(txt)
}

# y <- c(8, 8, 4, 3, 5, 6, 8, 3, 4, 5, 7, 10, 11)
# x <- seq_along(y)
# plot(x, y, type = "l")
# abline(h = threshold <- 5.5, col = 2)
# start <- 3
# end <- 10
# col <- "lightgrey"
intersection <- function(x, g, h, plot = F) {
  if(plot) {
    plot(x, g, type = "o")
    lines(x, h, type = "o", col = 4)
  }

  if(length(x) == 1) return(list(x = rep(x, 2), y = c(g, h)))

  # intersect two lines g and h
  dg <- g[1]
  dh <- h[1]

  kg <- diff(g[1:2])/diff(as.numeric(x))
  kh <- diff(h[1:2])/diff(as.numeric(x))

  ys <- (dg*kh - dh*kg) / (kh - kg)
  dx <- (ys - dg)/kg

  xs <- x[1] + dx
  if (plot) points(xs, ys, col = 2)

  return(list(x = xs, y = ys))
}

polygon_drought2 <- function(x, y, threshold, only.drought = T, col = "lightgrey") {
  x <- as.POSIXct(x)
  y <- as.vector(as.numeric(y))
  threshold <- as.vector(as.numeric(threshold))
  if(length(threshold) == 1) threshold <- rep.int(threshold, length(x))
  upper <- y >= threshold
  group <- .group(upper)

  event <- if(only.drought) group[!upper] else group
  poly <- list()
  for(i in as.numeric(as.character(unique(event)))){
    idx <- which(group == i)
    ep <- min(idx) - c(1, 0)
    inters1 <- intersection(x = as.POSIXct(x[ep]),
                            g = as.vector(y[ep]),
                            h = as.vector(threshold[ep]))

    ep <- max(idx) + c(0, 1)
    ep <- ep[ep<=length(x)]
    inters2 <- intersection(x = as.POSIXct(x[ep]),
                            g = as.vector(y[ep]),
                            h = as.vector(threshold[ep]))

    poly[[i]] <- list(x = c(inters1$x, as.POSIXct(x[idx]),
                            inters2$x, rev(as.POSIXct(x[idx]))),
                      y = c(inters1$y, as.vector(threshold[idx]),
                            inters2$y, rev(as.vector(y[idx]))))

  }

  u <- lapply(poly, polygon, col = col)
}



polygon_drought <- function(start, end, x, y, threshold = threshold, ...) {
  idx <- which(x > start & x <= end)
  idx <- c(idx, max(idx) + 1)
  poly <- list(x = as.POSIXct(c(start, x[idx], x[max(idx)], start)) - 3600,
               y = c(threshold, y[idx], threshold, threshold))

  # split polygon at threshold
  polygon(x = poly, ...)

  polygon(list(x = c(as.POSIXct(c(start, start, rep(x[max(idx)], 2))) - 3600),
               y = c(threshold, rep(max(y[idx], na.rm = T), 2), threshold)),
          col = "white", border = NA)
}

.circular_threshold <- function(x, fun, n = 31, ...) {
  x <- vary_threshold(unleap(x), "daily", fun, ...)
  x$threshold <- ma(x$threshold, n = n, sides = 2)
  x$day <- format(time(x), "%j")

  idx.bad <- which(is.na(as.vector(x$threshold)))
  day.bad <- as.vector(x$day[idx.bad])

  cc <- na.omit(x)

  x[idx.bad, "threshold"] <-  as.vector(cc[, "threshold"])[match(day.bad, cc$day)]

  return(x[, "threshold"])
}

.limits <- function(x) do.call(c, xts::.parseISO8601(x))


.plot_period <- function(values, period, ylim, threshold, rng = data.frame()) {
  x <- values[period, ]
  if(nrow(x) == 0) return()
  year <- format(xts::.parseISO8601(period)[[1]], "%Y")
  xlim <- as.POSIXct(paste0(year, c("-01-01", "-12-31")))
  plot(x, major.ticks = "month", type = "l", major.format = "%b",
       main = paste("Year", year), minor.ticks = T,
       xlim = xlim, ylim = ylim)


  if(!is.null(rng) && nrow(rng)) {
    rng <- unlist(rng[, c("start", "end")])
    rng <- as.Date(rng, origin = "1970-01-01")

    drought <- values[time(values) >= rng[1] -1 & time(values) <= rng[2] +1, ]
    polygon_drought2(x=time(drought), y=drought$discharge, threshold = threshold)

  }

  abline(h = threshold, col = 2)

  lines(.circular_threshold(values, quantile, probs = 0.5, na.rm = T),
        col = 4, lty =2)
  lines(.circular_threshold(values, quantile, probs = 0.2, na.rm = T),
        col = 2, lty = 2)
  # legend("topright", col = c(1, 2, 2, 4),
  #        legend = c("discharge", "Q80 const.",
  #                   "daily Q80 (n=31)", "daily Q50 (n=31)"),
  #        lty = c(1, 1, 2, 2))
}

.plot_period2 <- function(values, period, year, ylim, xlim, threshold,
                          rng = data.frame(), lwd) {
  x <- values[period, ]
  x <- data.frame(time = as.POSIXct(time(x)), value = as.vector(x))

  y <- x[x$time >= xlim[1] & x$time < xlim[2], ]

  months <- seq(from = xlim[1], length.out = 13, by = "months")


  plot.new()
  plot.window(xlim = xlim, ylim = ylim, xaxs = "i", yaxs = "i")

  if(!is.null(rng) && nrow(rng)) {
    rng <- unlist(rng[, c("start", "end")])
    rng <- as.Date(rng, origin = "1970-01-01")

    drought <- values[time(values) >= rng[1] -1 & time(values) <= rng[2] +1, ]
    polygon_drought2(x=time(drought), y=drought$discharge, threshold = threshold)

  }


  qmin <- which.min(y$value)
  abline(v = months, col = "lightgrey", lwd = lwd)
  points(x = y$time[qmin], y = y$value[qmin], pch = 21, bg = "coral")
  lines(x = y$time, y = y$value)



  axis(2, padj = 0.8, lwd = lwd)
  mtext(expression(Discharge ~ (m^scriptscriptstyle(3) ~ s^scriptscriptstyle(-1))),
        side = 2, line = 1.5, cex = 0.7)
  monthLab <- c("Jan", "Feb", "Mar", "Apr", "May", "Jun",
                "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")

  text(x = xlim[2], y = ylim[2], adj = c(1.2, 1.5), labels = year, font = 2, cex = 1.25)
  axis(side = 1, at = months, labels = F, lwd = lwd)
  axis(side = 1, at = head(months, -1) + 14 * 24 *3600 , labels = monthLab , tick = F, padj = -1.3, lwd = lwd)
  box(lwd = lwd)


  abline(h = threshold, col = 2, lwd = 0.25)

  lines(.circular_threshold(values, quantile, probs = 0.5, na.rm = T),
        col = 4, lty =2)
  lines(.circular_threshold(values, quantile, probs = 0.2, na.rm = T),
        col = 2, lty = 2)
}


plot_ts <- function(data = NULL, meta = compare_lf(data)) {


  if(!inherits(meta, "hpt")) stop("argument meta must of class xts")

  att <- attr(meta, "lf")

  if(!is.null(data)) {
    if(!inherits(data, "xts")) stop("argument data must be an xts time series")
    ninf <- xtsAttributes(data)$station
    nmeta <- att$station

    if(ninf != nmeta) stop("Arguments mismatch: ", ninf, ", ", nmeta)
  }

  op <- par(no.readonly = T)
  on.exit(par(op))

  values <- data
  ref <- values[att$refPeriod, ]
  periods <- lapply(att$periods, function(x) values[x, ])
  periods <- periods[sapply(periods, nrow) > 0]

  par(mfcol = c(length(periods) + 1, 1),
      oma = rep(0, 4), mar = c(3, 4, 2, 0) + 0.1)


  threshold <- att$threshold
  ylim <- c(0, quantile(as.vector(ref), 0.95, na.rm = T))

  plot(log10(ref), major.ticks = "year", type = "l", major.format = "%Y",
       main = "Reference Period", minor.ticks = F, ylab = "log discharge",
       xlim = .limits(att$refPeriod))#, ylim = ylim)
  abline(h = log10(threshold), col = 2)

  for(i in att$periods) {
    year <- format(xts::.parseISO8601(i)[[1]], "%Y")
    .plot_period(values = values, period = i, ylim = ylim, threshold = threshold,
                 rng = att$drought[[paste0("y", year)]])
  }


}



plot_ts2 <- function(data = NULL, meta = compare_lf(data), pp,
                     tree = NULL, year = 2015,
                     ylim.pp = NULL,
                     panel = c("monthly", "weekly", "daily")) {
  panel <- match.arg(panel)

  if(!inherits(meta, "hpt")) stop("argument meta must of class xts")

  att <- attr(meta, "lf")

  if(!is.null(data)) {
    if(!inherits(data, "xts")) stop("argument data must be an xts time series")
    ninf <- xtsAttributes(data)$station
    nmeta <- att$station

    if(ninf != nmeta) stop("Arguments mismatch: ", ninf, ", ", nmeta)
  }


  lwd <- 0.4


  ref <- data[att$refPeriod, ]
  threshold <- att$threshold
  ylim <- c(0, quantile(as.vector(ref), 0.95, na.rm = T))


  xlim <- as.POSIXct(paste0(as.numeric(year) + c(0, 1), c("-01-01")))

  par(mar = c(0.25, 3.5, 0.5, 0.1))
  if(!is.null(tree)) {
    .panel_trd(as.Date(xlim), tree)
  }

  if(!is.null(pp)) {
    pp$year <- as.numeric(format(pp$Date, "%Y"))
    pp <- pp[pp$year == year, c("Date", "value")]

    grp <- switch(panel,
                  monthly = as.numeric(format(pp$Date, "%m")),
                  weekly = as.numeric(format(pp$Date, "%W")),
                  daily = as.numeric(format(pp$Date, "%j")))

    sums <- unname(tapply(pp$value, grp, sum))
    .panel_pp(sums, lwd, ylim = ylim.pp)
  }

  par(mar = c(1.4, 3.5, 0.5, 0.1))
  .plot_period2(values = data, period = att$periods[paste0("y", year)],
                year = year, ylim = ylim, xlim = xlim, threshold = threshold,
                rng = att$drought[[paste0("y", year)]], lwd)


}

.panel_pp <- function(pp, lwd, ylim = NULL) {
  if(is.null(ylim)) ylim <- c(0, max(pp))

  #plot.new()
  barplot2(unname(pp), space = 0, col = "white", lwd = lwd,
           ylim = rev(ylim), xaxs = "i", yaxs = "i", axes = F)

  #box(lwd = lwd)
  axis(2, padj = 0.8, lwd = lwd)
  mtext("Precipitation", side = 2, line = 1.5, cex = 0.7)
}

barplot2 <- function (height, width = 1, space = NULL, names.arg = NULL,
                      legend.text = NULL, beside = FALSE, horiz = FALSE, density = NULL,
                      angle = 45, col = NULL, border = par("fg"), main = NULL,
                      sub = NULL, xlab = NULL, ylab = NULL, xlim = NULL, ylim = NULL,
                      xpd = TRUE, log = "", axes = TRUE, axisnames = TRUE, cex.axis = par("cex.axis"),
                      cex.names = par("cex.axis"), inside = TRUE, plot = TRUE,
                      axis.lty = 0, offset = 0, add = FALSE, args.legend = NULL,
                      lwd = 1, ...)
{
  if (!missing(inside))
    .NotYetUsed("inside", error = FALSE)
  if (is.null(space))
    space <- if (is.matrix(height) && beside)
      c(0, 1)
  else 0.2
  space <- space * mean(width)
  if (plot && axisnames && is.null(names.arg))
    names.arg <- if (is.matrix(height))
      colnames(height)
  else names(height)
  if (is.vector(height) || (is.array(height) && (length(dim(height)) ==
                                                 1))) {
    height <- cbind(height)
    beside <- TRUE
    if (is.null(col))
      col <- "grey"
  }
  else if (is.matrix(height)) {
    if (is.null(col))
      col <- gray.colors(nrow(height))
  }
  else stop("'height' must be a vector or a matrix")
  if (is.logical(legend.text))
    legend.text <- if (legend.text && is.matrix(height))
      rownames(height)
  stopifnot(is.character(log))
  logx <- logy <- FALSE
  if (log != "") {
    logx <- length(grep("x", log)) > 0L
    logy <- length(grep("y", log)) > 0L
  }
  if ((logx || logy) && !is.null(density))
    stop("Cannot use shading lines in bars when log scale is used")
  NR <- nrow(height)
  NC <- ncol(height)
  if (beside) {
    if (length(space) == 2)
      space <- rep.int(c(space[2L], rep.int(space[1L],
                                            NR - 1)), NC)
    width <- rep_len(width, NR)
  }
  else {
    width <- rep_len(width, NC)
  }
  offset <- rep_len(as.vector(offset), length(width))
  delta <- width/2
  w.r <- cumsum(space + width)
  w.m <- w.r - delta
  w.l <- w.m - delta
  log.dat <- (logx && horiz) || (logy && !horiz)
  if (log.dat) {
    if (min(height + offset, na.rm = TRUE) <= 0)
      stop("log scale error: at least one 'height + offset' value <= 0")
    if (logx && !is.null(xlim) && min(xlim) <= 0)
      stop("log scale error: 'xlim' <= 0")
    if (logy && !is.null(ylim) && min(ylim) <= 0)
      stop("log scale error: 'ylim' <= 0")
    rectbase <- if (logy && !horiz && !is.null(ylim))
      ylim[1L]
    else if (logx && horiz && !is.null(xlim))
      xlim[1L]
    else 0.9 * min(height, na.rm = TRUE)
  }
  else rectbase <- 0
  if (!beside)
    height <- rbind(rectbase, apply(height, 2L, cumsum))
  rAdj <- offset + (if (log.dat)
    0.9 * height
    else -0.01 * height)
  delta <- width/2
  w.r <- cumsum(space + width)
  w.m <- w.r - delta
  w.l <- w.m - delta
  if (horiz) {
    if (is.null(xlim))
      xlim <- range(rAdj, height + offset, na.rm = TRUE)
    if (is.null(ylim))
      ylim <- c(min(w.l), max(w.r))
  }
  else {
    if (is.null(xlim))
      xlim <- c(min(w.l), max(w.r))
    if (is.null(ylim))
      ylim <- range(rAdj, height + offset, na.rm = TRUE)
  }
  if (beside)
    w.m <- matrix(w.m, ncol = NC)
  if (plot) {
    dev.hold()
    opar <- if (horiz)
      par(xaxs = "i", xpd = xpd)
    else par(yaxs = "i", xpd = xpd)
    on.exit({
      dev.flush()
      par(opar)
    })
    if (!add) {
      plot.new()
      plot.window(xlim, ylim, log = log, ...)
    }
    xyrect <- function(x1, y1, x2, y2, horizontal = TRUE,
                       ...) {
      if (horizontal)
        rect(x1, y1, x2, y2, lwd = 1.5 * lwd)
      else {
        keep <- x1 != x2
        rect(y1[keep], x1[keep], y2[keep], x2[keep], lwd = 1.5 * lwd)
        segments(min(y1), min(x1), max(y2), min(x1), lwd = lwd)
      }
    }
    if (beside)
      xyrect(rectbase + offset, w.l, c(height) + offset,
             w.r, horizontal = horiz, angle = angle, density = density,
             col = col, border = border, lwd = 1.5 * lwd)
    else {
      for (i in 1L:NC) {
        xyrect(height[1L:NR, i] + offset[i], w.l[i],
               height[-1, i] + offset[i], w.r[i], horizontal = horiz,
               angle = angle, density = density, col = col,
               border = border, lwd = lwd)
      }
    }
    if (axisnames && !is.null(names.arg)) {
      at.l <- if (length(names.arg) != length(w.m)) {
        if (length(names.arg) == NC)
          colMeans(w.m)
        else stop("incorrect number of names")
      }
      else w.m
      axis(if (horiz)
        2
        else 1, at = at.l, labels = names.arg, lty = axis.lty,
        cex.axis = cex.names, ...)
    }
    if (!is.null(legend.text)) {
      legend.col <- rep_len(col, length(legend.text))
      if ((horiz & beside) || (!horiz & !beside)) {
        legend.text <- rev(legend.text)
        legend.col <- rev(legend.col)
        density <- rev(density)
        angle <- rev(angle)
      }
      xy <- par("usr")
      if (is.null(args.legend)) {
        legend(xy[2L] - xinch(0.1), xy[4L] - yinch(0.1),
               legend = legend.text, angle = angle, density = density,
               fill = legend.col, xjust = 1, yjust = 1)
      }
      else {
        args.legend1 <- list(x = xy[2L] - xinch(0.1),
                             y = xy[4L] - yinch(0.1), legend = legend.text,
                             angle = angle, density = density, fill = legend.col,
                             xjust = 1, yjust = 1)
        args.legend1[names(args.legend)] <- args.legend
        do.call("legend", args.legend1)
      }
    }
    title(main = main, sub = sub, xlab = xlab, ylab = ylab,
          ...)
    if (axes)
      axis(if (horiz)
        1
        else 2, cex.axis = cex.axis, ...)
    invisible(w.m)
  }
  else w.m
}

.panel_trd <- function(xlim, trd, lwd = 1) {
  ultrd <- trd[, setdiff(names(trd), "Date")]

  ylim <- c(0, max(ultrd, na.rm = T) * 1.1)

  plot.new()
  plot.window(xlim = xlim, ylim = rev(ylim), xaxs = "i", yaxs = "i")


  for(i in seq_len(ncol(ultrd))) {
    lines(trd$Date, ultrd[, i], type = "s", col = grey(1-1/i))
  }

  box(lwd = lwd)
  legend('bottom', legend = colnames(ultrd), col = grey(1-1/seq_len(i)),
         ncol = i, cex=0.7, bty='n', lwd = 1)
  axis(2, padj = 0.8, lwd = lwd)
  mtext("tree ring index", side = 2, line = 1.5, cex = 0.7)
}



is.leapyear <- function(year) {
  !(year %% 4) & as.logical(year %% 100) | !(year %% 400)
}

unleap <- function(x, skip = 366) {
  if(!inherits(time(x), "Date")) stop("unleap() only works with class 'Date'.")
  day <- as.numeric(format(time(x), "%j"))
  leap.year <- is.leapyear(as.numeric(format(time(x), "%Y")))

  y <- x[!leap.year | leap.year & day != skip, ]
  return(y)
}

leap <- function(x, skip = 366) {
  if(!inherits(time(x), "Date")) stop("unleap() only works with class 'Date'.")
  year <- as.numeric(unique(format(time(x), "%Y")))
  leapyear <- year[is.leapyear(year)]

  # class Date is zero indexed, Date(0) = "1970-01-01"
  # year has to be a leap year
  day <- format(as.Date(skip - 1, origin = "2000-01-01"), "-%m-%d")
  insert <- as.Date(paste0(leapyear, day))

  y <- approx(time(x), as.vector(x$discharge), xout = insert)
  y <- xts(data.frame(discharge = y$y), order.by = y$x)

  return(rbind(x, y))
}



# .trySilent <- function(expr, ..., finally) {
#   tryCatch(expr = expr, ..., finally = finally, error = function(e) numeric())
# }

.trySilent <- function(expr, ..., finally) {
  eval(expr)
}






.warn <- function(x, what = "???") {
  failed <- which(sapply(x, length) == 0)
  if(length(failed)) {
    warning("***   ", what, " failed for ", length(failed), " stations: ",
            paste(failed, collapse = ", "), call. = FALSE)
  }
}


run_country <- function(outdir = ".", save = TRUE, parallel = TRUE, ...) {

  infile <- readRDS(file.path(outdir, "infile.rds"))

  if(parallel) {
    registerDoParallel()
    result <- foreach(i=infile) %dopar% compare_lf(i, ...)
  } else {
    result <- lapply(infile, compare_lf, ...)
  }
  .warn(result, what = "Processing full time series")


  trim <- c(y2015 = "2015-01-01::2015-10-31", y2003 = "2003-01-01::2003-12-31")
  if(parallel) {
    result_trimmed <- foreach(i=infile) %dopar% compare_lf(i["::2015-10-31", ],
                                                           periods = trim, ...)
  } else {
    result_trimmed <- lapply(infile, function(i) compare_lf(i["::2015-10-31", ],
                                                            periods = trim, ...))
  }
  .warn(result_trimmed, what = "Processing trimmed time series")


  if(save) {
    saveRDS(file = file.path(outdir, "result.rds"), result)
    saveRDS(file = file.path(outdir, "result_trimmed.rds"), result_trimmed)
    save(result, result_trimmed, infile,
         file = file.path(outdir, "results.RData"))
  } else {
    return(result)
  }
}
