### wrapper to import all files of a country ----

import_country <- function(station.info, inpdir = "./discharge/", outdir = ".",
                           save = TRUE,  parallel = TRUE, ...) {

  filenames <- file.path(inpdir, station.info$filename)

  if(parallel) {
    registerDoParallel()
    infile <- foreach(i=filenames) %dopar% import(i,  station.info = station.info, ...)
  } else {
    infile <- lapply(filenames, function(x)
      .trySilent(import(x, station.info = station.info, ...)))
  }


  unused <- setdiff(list.files(inpdir), station.info$filename)
  if(length(unused)) {
    warning("There are ", length(unused), " unused files in ", shQuote(inpdir), ": ",
            paste(shQuote(unused), collapse = ", "), call. = FALSE)
  }

  .warn(infile, what = "Import")

  if(save) {
    saveRDS(file = file.path(outdir, "infile.rds"), infile)
  } else {
    return(infile)
  }
}

.infoToAttrib <- function(x) {
  y <- as.list(x)
  y[["station"]] <- y[["name"]]

  cols <- c("x2", "y2", "z")
  idx <- match(c("x", "y", "z"), substring(names(y), 0L, 1L))
  if(sum(!is.na(idx))) y[["coordinates"]] <- unlist(x[1, idx])
  return(y)
}


### general import function ----
import <- function(file, fun = read.csv, station.info,
                   format = "%Y-%m-%d", ...) {

  # station info must no have cols of class factor
  i <- sapply(station.info, is.factor)
  station.info[i] <- lapply(station.info[i], as.character)

  # read discharge file and trim leading/trailing missing observations
  csv <- fun(file = file, ..., as.is = TRUE)
  trimmed <- na.trim(csv)

  delta <- nrow(csv) - nrow(trimmed)
  if(delta) warning(basename(file), ": Trimmed ", delta,
                    " leading/trailing missing observations.", call. = FALSE)

  if(ncol(trimmed) != 2) {
    stop(basename(file), ": fun() must return a data.frame with two columns.",
         call. = FALSE)
  }

  time <- as.Date(trimmed[, 1], format = format)
  discharge <- as.numeric(trimmed[, 2])

  dups <- which(duplicated(time))
  if(length(dups)) warning(basename(file), ": duplicated time indices found (",
                           paste(shQuote(unique(time[dups])), collapse = ", "),
                           ")", call. = FALSE)
  nas <- is.na(time)
  if(any(nas)) {
    warning(basename(file), ": ", sum(nas), " NA values in time index",
            call. = FALSE)
  }

  y <- xts(x = data.frame(discharge = discharge), order.by = time)
  padded <- regularize(y)
  delta <- nrow(padded) - nrow(y)
  if(delta) warning(basename(file), ": Padded with ", delta,
                    " NA values to obtain a regular time series.", call. = FALSE)

  # interpolate NA values if gap is of length one
  padded <- .fill.na(padded, max.len = 1)


  # check if fun() has meta data stored as attribute
  meta <- attr(csv, "meta")

  # if incomplete, get missing attribs from station.info
  if(!all(unlist(.meta) %in% names(meta))) {

    # check if there is a corresponding entry in station.info
    idx <- which(basename(file) == station.info$filename)[1]
    if(is.na(idx)) {
      warning("no entry in station.info for station ", file, call. = F)
      return(NULL)
    }

    meta.info <- .infoToAttrib(station.info[idx, ])

    # TODO: check for conflicting values

    # meta info from station.info wins
    meta <- c(meta.info, meta[setdiff(names(meta), names(meta.info))])
  }

  # meta data computed from time series of discharge
  meta.input <- list(time = range(time[!is.na(discharge)]),
                     data.coverage = sum(is.na(padded) / nrow(padded)))

  # finally, only keep required and optional arguments
  # as set in global variable .meta
  meta <- meta[unlist(.meta)]
  meta <- c(meta, meta.input)

  meta$coord.local <- c(meta$coordinates, "epsg" = meta$epsg)

  # transform coordinate reference system
  if(!length(meta$epsg) || !is.finite(meta$epsg)) {
    stop("No or invalid epsg code specified for ", shQuote(file))
  }
  xy <- transform_crs(as.list(meta$coordinates[c("x", "y")]), from = meta$epsg)
  names(xy) <- c("x2", "y2")
  meta$coordinates <- c(unlist(xy), meta$coordinates["z"])
  meta$epsg <- 3034

  xtsAttributes(padded) <- meta[sapply(meta, length) > 0]
  .check_attr(padded)

  return(padded)
}


### Belgium / Flanders ----
read.csv.be <- function(file, ...) {

  txt <-  readLines(file, warn = FALSE)

  lines.header <- grep("Timestamp;Value;Quality", head(txt, 50), fixed = T) - 1
  header <- head(txt, lines.header)

  infile <- read.csv2(text = txt, header = T, skip = lines.header,
                      colClasses = c("character", rep("numeric", 2), rep("NULL", 2)),
                      col.names = c("time", "value", "flag", NA, NA),
                      strip.white = TRUE, as.is = TRUE)

  time <- gsub(":00$|00.000", "00", infile$time)
  infile$time <- as.POSIXct(time, format = "%Y-%m-%dT%H:%M:%S%z")

  return(infile[, c("time", "value")])
}


###  Norway ----
read.dat.no <- function(file, ...) {

  infile <- read.table(file, col.names = c("y", "m", "d", "value"),
                       colClasses = "numeric", na.strings = "-9999.000000", ...)

  time <- as.Date(do.call(paste, c(as.list(infile[, c("y", "m", "d")]),
                                   sep = "-")))

  return(data.frame(time = time, value = infile$value))
}


# UK ----

read.nrfa <- function(file, ...) {
  txt <-  readLines(file)

  lines.header <- grep("data,last,", head(txt, 50), fixed = T)
  header <- head(txt, lines.header)

  infile <- read.csv(text = txt, header = F, skip = lines.header,
                     col.names = c("time", "value", "flag"),
                     colClasses = c("Date", "numeric", "NULL"),
                     strip.white = TRUE, as.is = TRUE)

  .parse_header_nrfa <- function(x) {

    y <- read.csv(text = x, header = F, strip.white = TRUE,
                  col.names = c("key", "component", "value"),
                  colClasses = c("character"), ...)

    meta <- list()
    for(i in seq_len(nrow(y))) {
      meta[[y$key[i]]][[y$component[i]]] <- y$value[i]
    }

    return(lapply(meta, as.list))

  }

  y <- infile[, c("time", "value")]


  meta <- .parse_header_nrfa(header)

  attr(y, "meta") <- list(
    eid = meta$station$id,
    station = sub(" at .+$", "", meta$station$name),
    river = sub("^.+ at ", "", meta$station$name),
    unit = sub("3", "^3", meta$dataType$units, fixed = TRUE),
    institution = meta$database$name,
    country = "gb",
    time = as.Date(unlist(meta$data[c("first", "last")], use.names = F)))

  return(y)
}


### Spain ----

read.txt.es <- function(file, ...) {

  infile <- read.table(file, colClasses = c("numeric"),
                       col.names = c("eid", "day", "month", "year", "value"),
                       na.strings = c("-100"), ...)

  time <- as.Date(do.call(paste, c(as.list(infile[, c("year", "month", "day")]),
                                   sep = "-")))

  return(data.frame(time = time, value = infile$value))
}


### Germany / Bavaria ----

read.gkd <- function(file, ...) {
  txt <-  readLines(file)

  lines.header <- grep("Datum;Mittelwert;Maximum;Minimum;", head(txt, 50),
                       fixed = T)
  header <- head(txt, lines.header - 3)

  infile <- read.csv2(text = txt, header = FALSE, skip = lines.header,
                      col.names = c("time", "mean", "max", "min", "flag"),
                      colClasses = c("Date", rep("numeric", 3), "character"),
                      strip.white = TRUE, ...)

  y  <- data.frame(time = infile$time, value = infile$mean)

  .parse_header_gkd <- function(x) {

    y <- read.csv2(text = x, header = F, strip.white = TRUE, as.is = TRUE,
                   col.names = c("key", "value"),
                   colClasses = c("character"))

    y$key <- sub("\\.?:$", "", y$key)

    xy <- y$value[y$key == "Koordinaten"]
    xy <- as.numeric(strsplit(xy, "[^[:digit:].]+")[[1]][2:4])
    epsg <- xy[3]
    xy <- xy[1:2]
    names(xy) <- c("x", "y")
    z <- y$value[y$key == "Pegelnullpunktsh\u00f6he"]
    z <- as.numeric(sub(",", ".", gsub("[^[:digit:],]+", "", z)))

    res <- list(institution = y$value[y$key == "Quelle"],
                station = y$value[y$key == "Messstellen-Name"],
                eid = y$value[y$key == "Messstellen-Nr"],
                river = y$value[y$key == "Gew\u00e4sser"],
                region = "Bayern",
                coordinates = c(unlist(xy), z = z),
                epsg = epsg)

    return(res)
  }

  attr(y, "meta") <- c(.parse_header_gkd(header),
                       list(unit = "m^3/s", country = "de"))

  return(y)
}
