classify_rp <- function(x, breaks = c(1, 1/0.9, 1/0.8, 2, 5, 20, 50, Inf)) {
  cut(x, breaks = breaks, include.lowest = TRUE)
}

regularize <- function(x, interval = "day") {
  x <- x[!is.na(time(x)), ]

  fullseq <- seq(from = min(time(x)), to = max(time(x)), by = interval)
  missing <- fullseq[!fullseq %in% time(x)]

  if(length(missing)) {
    gaps <- xts(x = data.frame(discharge = rep_len(NA_real_, length(missing))), order.by = missing)
    x <- rbind(x, gaps)
  }

  return(x)
}

.year <- function(x) as.numeric(format(as.Date(x), "%Y"))
.dayofyear <- function(x) as.numeric(format(as.Date(x), "%j"))

.tapply.ind <- function(x, index, fun, drop = F, ...){
  mask <- function(x, ...) sapply(x == fun(x, ...), isTRUE)
  grp <- lapply(split(x, index, drop = drop), mask, na.rm = T)

  return(unsplit(grp, index, drop = drop))
}


.am  <- function(y, choose.first = TRUE, name = "") {
  rng <- as.numeric(format(range(time(y)), "%Y"))
  group <- .year_as_factor(time(y), rng[1], rng[2])

  # a warning is given if a year contains only NA values
  idx <- suppressWarnings(.tapply.ind(as.vector(y), group, min, na.rm = TRUE))
  day <- split(seq_along(idx)[idx], group[idx], drop = F)
  day.of.year <- tapply(idx, group, which)

  # only warn if minima were observed in different events
  dup <- names(which(sapply(day, function(x) length(unique(diff(x))) > 1)))
  if(length(dup)) {
    warning(paste0(name, ": For the year(s) '", paste(dup, collapse = ", "),
                   "' the minimum discharge was observed on more than one day."),
            call. = F)
  }

  if (choose.first) {
    day <- lapply(day, function(x) x[1])
    day.of.year <- lapply(day.of.year, function(x) x[1])
  }

  day <- na.omit(unlist(day, use.names = FALSE))
  day.of.year <- na.omit(unlist(day.of.year, use.names = FALSE))

  return(cbind(y[day, ], day = day.of.year))
}

.rp_ev <- function(x, fit) {
  dist <- names(fit$parameters)[1]
  cdf <- match.fun(paste0("cdf", dist))
  prob <- cdf(x = x, para = fit$parameters[[1]])
  prob <- prob + fit$freq.zeros * (1 - prob)

  if(fit$extreme == "maximum")  {
    prob <- ifelse(x == 0, 0, prob)
    prob <- 1 - prob
  } else {
    prob <- ifelse(x == 0, fit$freq.zeros, prob)
  }

  rp <- 1 / prob

  return(rp)
}


# write a Date method for .circular_mean
.circular_mean <- function(x, max.val = 365) {
  ang <- x/max.val * 2 * pi
  xy <- cbind(x = cos(ang), y = sin(ang))
  avg <- colMeans(xy)
  res <- c("value" = (atan2(avg["y"], avg["x"]) / 2 / pi * max.val) %% max.val,
           "var" = sqrt(sum(avg^2)))
  return(res)
}



.year_as_factor <- function(x, from, to) {
  x <- as.Date(x)
  return(factor(as.numeric(format(x, "%Y")), seq(from, to)))
}


test_crs <- function(gauge, crs, f = 1) {
  xy <- transform_crs(x = gauge[, c("x", "y")], from = crs, to = CRS("+init=epsg:3034"))
  x <- data.frame(x2 = xy[, 1], y2 = xy[, 2])
  #data("country")

  require(ggplot2)

  lim <- list(data = list(x = extendrange(x$x2, f = f),
                          y = extendrange(x$y2, f = f)),
              europe = list(x = .bb[c("xmin", "xmax")],
                            y = .bb[c("ymin", "ymax")]))
  limit <- lim[["data"]]

  ggplot2::ggplot() +
    ggplot2::geom_polygon(data = country, aes(x2, y2, group = group),
                          col = "#525252", fill = "#ffffff") +
    ggplot2::coord_fixed(xlim = limit[["x"]], ylim = limit[["y"]]) +
    ggplot2::geom_point(data = x, ggplot2::aes(x2, y2)) +
    ggplot2::theme_bw(10)+
    ggplot2::theme(axis.line=ggplot2::element_blank(),
                   axis.text.x=ggplot2::element_blank(),
                   axis.text.y=ggplot2::element_blank(),
                   axis.ticks=ggplot2::element_blank(),
                   panel.grid = ggplot2::element_blank(),
                   axis.title.x=ggplot2::element_blank(),
                   axis.title.y=ggplot2::element_blank(),
                   plot.background=ggplot2::element_blank(),
                   strip.background=ggplot2::element_rect(fill = "white"),
                   panel.background = element_rect(fill = "#C5DAE3"))
}


correct_crs <- function(x, from, europe = CRS("+init=epsg:3034"), gauge,
                        k = 1, d = 0) {

  att <- attr(x, "lf")
  xy <- att$coordinates[1:2]

  dist <- as.matrix(gauge[, names(xy)]) -
    matrix(unlist(xy), nrow = nrow(gauge), ncol = 2, byrow = T)
  dist <- rowSums(abs(dist))
  idx <- which.min(dist)
  if(dist[idx] > 10) stop("deviation:, ", dist[idx])

  # try new one
  new <- transform_crs(gauge[idx, c("x", "y")] * k + d, from = from,
                       to = europe, suffix = "2")

  attr(x, "lf")$coordinates <- as.data.frame(c(new, gauge[idx, "z", drop = F]))

  return(x)
}

# convert coordinates
transform_crs <- function(x, y = NULL, suffix = "2", from,
                          to = CRS("+init=epsg:3034")) {

  xyl <- xy.coords(x, y)
  names <- xyl[c("xlab", "ylab")]
  unnamed <- sapply(names, is.null)
  names[unnamed] <- c("x", "y")[unnamed]

  if (is.numeric(from)) {
    from <- lapply(unique(from), function(x) CRS(paste0("+init=epsg:", x)))

  }

#   from <- lapply(from, function(x) CRS(paste0("+init=epsg:", x)))
#   res <- mapply(FUN = transform_crs, as.list(x, y = vector("list", length(x[[1]])),
#                                              suffix = suffix, from = from, to = to)
#
#                 return(do.call(rbind, res))


  xy.new <- data.frame(matrix(NA, length(xyl[[1]]), ncol = 2,
                              dimnames = list(NULL, paste0(names, suffix))))

  xy <- as.data.frame(xyl[c("x", "y")])
  complete <- complete.cases(xy)
  xy <- xy[complete, ]

  coordinates(xy) <- c("x", "y")
  proj4string(xy) <- from[[1]]

  xy.new[complete, ] <- as.data.frame(spTransform(x = xy, CRSobj = to))
  return(xy.new)
}

.coverage_station <- function(x) {
  att <- .extract_value(x, key = c("region", "station", "coordinates"))
  if(is.na(att$station)) att$station <- paste(att$coordinates[1:2], collapse = " - ")
  att <- att[setdiff(names(att), "coordinates")]


  mm <- attr(x, "lf")$mm7
  time(mm) <- as.Date(format(time(mm), "%Y-%m-01"))
  mm[!is.finite(mm)] <- NA

  fullseq <- seq(min(time(mm)), max(time(mm)), by = "months")
  gap <- fullseq[!fullseq %in%  time(mm)]
  if(length(gap)) {
    mm <- rbind(mm, xts(data.frame(rep(NA_real_, length(gap))), order.by = gap))
  }

  f <- .group(as.vector(complete.cases(mm)))
  coverage <- lapply(split(mm, f, drop=T),
                     function(x) data.frame(start=min(time(x)),
                                            end=max(time(x)),
                                            value =!any(is.na(x))))
  coverage <- cbind(do.call(rbind, coverage), att)

  coverage[!coverage$value, "start"] <- .extend_month(coverage[!coverage$value, "start"], 1)
  coverage[!coverage$value, "end"] <- .extend_month(coverage[!coverage$value, "end"], -1)

  return(coverage)
}

.extend_month <- function(x, n = 1) {
  y <- x
  for(i in seq_along(y)) {
    y[i] <- seq(from = x[i], by = paste(n, "months"), length.out = 2)[2]
  }

  return(y)
}

# data coverage
coverage <- function(x) {
  require(ggplot2)

  coverage <- do.call(rbind, lapply(x, .coverage_station))
  coverage$station <- substr(coverage$station, start = 0L, stop = 20L)

  coverage$value <- factor(coverage$value, levels = c(FALSE, TRUE))

  ggplot(coverage) +
    geom_segment(aes(x=start, xend=end, y=station, yend=station,
                     col=value), size=2) +
    coord_cartesian(xlim = as.Date(c("1970-01-01", "2016-01-01"))) +
    scale_color_discrete("data coverage", drop = FALSE) +
    facet_grid(region~., scales = "free_y", space = "free_y") +
    theme(axis.title=element_blank()) +
    theme_bw(10) +
    theme(legend.position = "none",
          axis.title.x=element_blank(),
          axis.title.y=element_blank())
}



# classify values due to their neighbours
.group <- function(x) {
  grp <- c(0, cumsum(diff(as.numeric(x)) != 0))

  if(grp[1] == 0) grp <- grp + 1

  return(factor(grp))
}

get_attr_infile <- function(x) {
  att <- xtsAttributes(x)
  meta <- att[setdiff(names(att), c("time", "coordinates"))]
  coord <- att[["coordinates"]]
  meta[names(coord)] <- coord
  time <- att[["time"]]
  meta[names(time)] <- time

  return(meta)
}

get_attr_result <- function(x) {
  att <- attr(x, "lf")

  coord <- attr(x, "lf")[["coordinates"]]
  att[names(coord)] <- coord

  return(att)
}

get_attr <- function(x, drop.complex = F) {
  if (inherits(x, "xts")) {
    y <- get_attr_infile(x)
  } else if(inherits(x, "hpt")) {
    y <- get_attr_result(x)
  } else stop("get_attr does not know how to handle an object of class ",
                 class(x), call. = FALSE)

  if (drop.complex)  y <- y[sapply(y, function(x)
    !is.list(x) && is.vector(x) && length(x) == 1)]

  return(y)
}

set_attr_infile <- function(x, attr) {
  xtsAttributes(x)[names(attr)] <- as.list(attr)

  return(x)
}

set_attr_result <- function(x, attr) {
  attr(x, "lf")[names(attr)] <- as.list(attr)

  return(x)
}

set_attr <- function(x, attr) {
  if (inherits(x, "xts")) {
    return(set_attr_infile(x, attr))
  } else if(inherits(x, "hpt")) {
    return(set_attr_result(x, attr))
  } else stop("set_attr does not know how to handle an object of class ",
                 class(x), call. = FALSE)
}

extract_metadata <- function(x) {
  gauge <- do.call(rbind.fill, lapply(x, function(x)
    as.data.frame(get_attr(x, drop.complex = T), stringsAsFactors = F)))

  rownames(gauge) <- NULL
  return(gauge)
}


update_metadata <- function(x, meta, attr = NULL) {
  if(!is.null(attr)) {
    y <- x
    attr <- intersect(colnames(meta), attr)

    oldval <- get_attr(y)
    id <- c("x2", "y2")
    id.old <- do.call(paste, c(oldval[id], sep = "-"))
    id.new <- do.call(paste, c(meta[, id], sep = "-"))

    newval <- meta[match(id.old, id.new), attr, drop = F]

    if(nrow(newval) != 1)  {
      stop("x, y coordinates not unique")
    } else {
      newval <- as.vector(newval)[attr]
    }

    return(set_attr(y, newval))
  }
}



.simpleCap <- function(x) {
  s <- strsplit(x, " ")[[1]]
  paste(toupper(substring(s, 1, 1)), substring(s, 2),
        sep = "", collapse = " ")
}


.ordinal.suffix <- function(x) {
  unit.pos <- x %% 10
  decade <- (x %% 100) %/% 10

  if(unit.pos %in% 1:3 && decade != 1) c("st", "nd", "rd")[unit.pos] else "th"
}

.date_english <- function(x) {
  if(is.character(x) || is.numeric(x)) {
    x <- as.Date(as.character(x), format = "%j")
  } else {
    x <- as.Date(x)
  }

  month <- month.abb[as.numeric(format(x, format = "%m"))]
  day  <- as.numeric(format(x, format = "%d"))

  return(paste0(month, " ", day, .ordinal.suffix(day)))
}


extract_timings <- function(x) {
  require(reshape2)
  att <- attr(x, "lf")

  y <- matrix(NA, nrow = nrow(x$values), ncol = 4,
              dimnames = list(period = rownames(x$values),
                              variable = c("start", "qmin", "vmin", "end")))
  y[, "qmin"] <- x$values[, "day"]
  y["reference", c("start", "end")] <- att$timing[, "reference"]

  per <- setdiff(rownames(x$values), "reference")

  y[per, c("start", "end")] <- t(apply(att$timing[, per], 2, .dayofyear))
  y["reference", c("start", "end")] <- att$timing[, "reference"]

  y[, "vmin"] <- (y[, "start"] + round(x$value[, "duration"], 1) - 1 ) %% 365

  y <- melt(y)
  cols <- c("x2", "y2", "station", "river", "country")
  gauge <- get_attr(x, drop.complex = T)[cols]

  return(cbind(y, gauge))
}

anomaly <- function(x) {
  timing <- do.call(rbind, lapply(x, extract_timings))

  # ensure that differences are smaller than a year
  .nd <- function(x) {
    y <- x %% 365
    ifelse(y > 365 /2, y - 365, y)
  }

  anomaly <- ddply(timing, .(x2, y2, station, river, country, variable), function(x) {
    #if(nrow(x) > 3) stop("too many observations")
    data.frame(value = .nd(c(x$value[x$period == "y2015"] - x$value[x$period == "reference"],
                             x$value[x$period == "y2015"] - x$value[x$period == "y2003"],
                             x$value[x$period == "y2003"] - x$value[x$period == "reference"])),
               period = factor(c("y2015 - reference", "y2015 - 2003", "y2003 - reference")))
  })

  anomaly$cl <- cut(anomaly$value, breaks = c(-182, -50, -30, -10, 10, 30, 50, 182))

  return(anomaly)
}


rbind.fill <- function(...) {
  x <- list(...)
  cols <- unique(unlist(lapply(x, names)))
  skel <- as.list(rep_len(NA, length(cols)))
  names(skel) <- cols

  .append_missing <- function(x, y) {
    y <- as.data.frame(y)
    missing <- setdiff(colnames(y), colnames(x))
    cbind(x, y[, missing])
  }

  y <- lapply(x, .append_missing, y = skel)

  return(do.call(rbind, y))
}



.fill.na <- lfstat::fill_na
